# Device Tree Essentials

## Device Tree

> Device Tree is a separate data structure for describing hardware.

\subsection{Motivation}
## Device Tree - Motivation #1

* On systems without device tree, hardware structure is "hard coded" and compiled as a
  part of the Linux kernel (board files)
* Each board has its own *board file* that defines and creates
  devices. It is a part of the kernel though.
* Re-compilation of kernel when hardware description changes.
  Kernel won't boot on another platform
* Bootloader loads single image (+initrd)

## Device Tree - Motivation #2

* Device tree is compiled separately, the resulting binary is
  called **device tree blob** or **flattened device tree** (FDT)
* One kernel can run on multiple platforms within same architecture
* In Linux kernel tree at `arch/<arch>/boot/dts`
* Custom device tree implementation also in non-GPL OS (e.g. GreenHills Integrity)

## Boot with Device tree

* The device tree blob and kernel is loaded from file (SD, tftp, ...) into RAM

  ```sh
  fatload mmc 0 0x8800000 devtree.dtb
  fatload mmc 0 0x0800000 uImage

  ```
* Bootloader may modify the blob (memory fixups, chosen node)
* Extract the kernel from uImage and boot it:

  ```sh
  bootm 0x0800000 - 0x8800000
  ```

* Kernel expands the blob to its internal representation
  called Expanded Device Tree (EDT)

\subsection{Structure}
## Device Tree - Structure and syntax

* Tree structure with named nodes
* C-like syntax (can be preprocessed by GCC)
* Each node can have an arbitrary number of named properties
* `compatible` property is a special property that links a node
  to a kernel driver
* Nodes can be linked to each other by `phandles`
* Nodes can have labels that will be expanded to
  node's phandle or path when referenced
* OS-specific or vendor-specific properties, nodes and
  compatible strings contain
  the os/vendor name as a prefix
* Formed by *Open Firmware* project, currently maintained by *devicetree.org* community

## Device Tree - Vendor prefix

```C
  linux,cma {
			compatible = "shared-dma-pool";
			linux,cma-default;
			...
		};

  dma-channel@80400030 {
				compatible = "xlnx,axi-dma-s2mm-channel";
				dma-channels = <0x01>;
				interrupts = <0x00 0x1d 0x04>;
				xlnx,datawidth = <0x20>;
				xlnx,device-id = <0x00>;
			};
```

\subsection{Properties and syntax}
## Device Tree - Property types

 Type        | Example
-------------|-----------------------------------
 empty value | `interrupt-controller;`
 u32 integer | `value = 0x11223344;`
 u64 integer | `value = <0x11223344 0x55667788>;`
 array       | `reg = <0xC1000 0x1000 0xA7000 0x1000>;`
 string      | `compatible = "prefix,the-string";`
 string list | `clk-names = "clk_per", "clk_phy"`;
 phandle     | `parent = <&another_node>;`

## Device Tree - Example #1

```C
	pcf8575: gpio@20 {
		compatible = "nxp,pcf8575";
		reg = <0x20>;
		interrupt-parent = <&irqpin2>;
		interrupts = <3 0>;
		gpio-controller;
		interrupt-controller;
	};
```

## Device Tree - Example #2

```C
/ {
	compatible = "accton,wr6202", "ralink,rt3052-soc";

	chosen {
		bootargs = "console=ttyS0,115200";
	};

	gpio-leds {
		compatible = "gpio-leds";

		wps {
			gpios = <&gpio0 14 GPIO_ACTIVE_LOW>;
		};
	};
...

```

## Device Tree - Build
\subsection{Device Tree Build}

* Device tree compiler - **dtc**
* Preprocessing (if required) done by any GCC-like preprocessor
* The blob can be converted back to its source form (debugging)
* **device tree blob is platform independent**

### Create a device tree blob from dts

`dtc -I dts -O dtb -o mx6ulp.dtb mx6ulp.dts`

### Create a device tree source from the blob

`dtc -I dtb -O dts -o mx6ulp.dts mx6ulp.dtb`

## Device Tree in kernel module #1

```C
// drivers/gpio/gpio-dwapb.c

static const struct of_device_id dwapb_of_match[] = {
	{ .compatible = "snps,dw-apb-gpio",
	  .data = (void *)0},
	{ .compatible = "apm,xgene-gpio-v2",
	  .data = (void *)GPIO_REG_OFFSET_V2},
	{ /* Sentinel */ }
};
MODULE_DEVICE_TABLE(of, dwapb_of_match);
```

\subsection{DevTree in kernel}
## Device Tree in kernel module #2

```C
// drivers/gpio/gpio-dwapb.c

static struct platform_driver dwapb_gpio_driver = {
	.driver		= {
		.name	= "gpio-dwapb",
		.pm	= &dwapb_gpio_pm_ops,
		.of_match_table = of_match_ptr(dwapb_of_match),
		.acpi_match_table = ACPI_PTR(dwapb_acpi_match),
	},
	.probe		= dwapb_gpio_probe,
	.remove		= dwapb_gpio_remove,
};

module_platform_driver(dwapb_gpio_driver);
```

\subsection{Tips}
## Device Tree - tips

* A long include chain is a common source of errors
* Inspect changes during the `DT Lifecycle`
  * preprocessing
  * build
  * dtb $\rightarrow$ FDT
  * FDT $\rightarrow$ EDT
  * Convert *dtb* back to *dts*
* Add structure-checking functionality to `probe()` function of your driver

## Device Tree in /proc

* EDT structure exported to filesystem
* make sure that `CONFIG_PROC_FS` and `CONFIG_OF` is enabled
* Each curly brace in device tree is a folder in `/proc/device-tree`
* accessible by dtc: `dtc -I fs ...`

## Device Tree in /proc

![](img/devtree_proc.png){width=80%}

## Device Tree tools

### dt_to_config
`/scripts/dtc/dt_to_config`

* Check device tree source against kernel configuration
* Find nodes that do not have drivers present or set for build
* `dt_to_config <path_to_dts_or_dtb>`

### dtdiff
`/scripts/dtc/dtxdiff`

* Compare two versions of DeviceTree (any format)
* Use `dtxdiff <dts file> <dtb file>`

## Runtime debugging

`Documentation/dynamic-debug-howto.txt`

* Enable debug for a specific file/line/function/module
* Enable kernel config `CONFIG_DYNAMIC_DEBUG`
* At boot-time - Add a query to kernel cmdline

  ```
  dyndbg="func bus_add_driver +p"
  dyndbg="func really_probe +p"
  ```

* At run-time - via debugfs

  ```bash
  echo "func bus_add_driver +p" > /sys/kernel/debug/dynamic_debug/control
  echo "func really_probe +p" > /sys/kernel/debug/dynamic_debug/control
  ```

## Runtime debugging

```
$ dmesg
bus: 'usb': really_probe: probing driver usb with device 4-2
bus: 'usb': really_probe: bound device 4-2 to driver usb
bus: 'usb': add driver r8152
bus: 'usb': really_probe: probing driver r8152 with device 4-2:2.0
r8152: probe of 4-2:2.0 rejects match -19
usbcore: registered new interface driver r8152
bus: 'usb': really_probe: probing driver r8152 with device 4-2:1.0
bus: 'usb': add driver cdc_ether
usbcore: registered new interface driver cdc_ether
usb 4-2: reset SuperSpeed Gen 1 USB device number 7 using xhci_hcd
r8152 4-2:1.0 eth0: v1.10.10
bus: 'usb': really_probe: bound device 4-2:1.0 to driver r8152
r8152 4-2:1.0 enp0s20u2: renamed from eth0
```

\subsection{Library}
## Device Tree - library

* [Device Tree for dummies](https://elinux.org/images/f/f9/Petazzoni-device-tree-dummies_0.pdf)
* [Device Tree Reference](https://elinux.org/Device_Tree_Reference)
* [devicetree.org](https://www.devicetree.org/)
* [Debugging devtree #1](https://elinux.org/images/0/04/Dt_debugging_elce_2015_151006_0421.pdf)
* [Debugging devtree #2](https://elinux.org/images/5/50/Dt_debugging_part_2.pdf)
