# Kernel modules

## What is a kernel module


> Each piece of code that can be added to the kernel at runtime is called a module.

## What is a kernel module

* Object code that is not linked into complete executable
* Allows extending the kernel functionality without
  a need to reboot the system - easy driver development.
* Same code can be built into the kernel (available from early boot), or as a module.
* Modules are stored as separate files in a filesystem. It needs to be
  mounted before modules can be loaded.
* Thanks to loadable modules, Linux kernel binary can be very small
  yet universal and multi-platform.

\subsection{Module vs application}
## Module vs applications

### Applications
* Event-driven or procedural
* Linked against external libraries
* Running in non-privileged mode
* Error *may not* cause system crash
* May be reentrant

### Modules
* Strictly event-driven
* Can use only functions exported from kernel.
* Running in privileged mode
* Error may cause system crash
* Must be reentrant
* Can export symbols to be used by other loadable module

## Simple kernel module

![](img/module_sample.png)

## Kernel module licensing

* Kernel provided under terms of GPL-2.0
* Licensing interface between userspace and kernel are syscalls
* Modules should be tagged by `MODULE_LICENSE` macro which specifies
  whether the module shall be linked with other modules
* Userspace headers are exception since they have to be included both to
  GPL kernel and (possibly) non-GPL user programs

```
/* SPDX-License-Identifier: GPL-2.0 WITH
   Linux-syscall-note */
```

## Userspace vs kernelspace

* **Kernelspace** - Linux kernel monolith and modules. Runs in privileged mode.
* **Userspace** - Other applications, running in non-privileged mode.
* Userspace applications perform privileged operations indirectly via kernel (syscalls)[^1]
  * Stable interface (only new syscalls added)
  * Part of std. C library

```
open, read, write, close, fsync, access, bind,
chown, chroot, ...
```

[^1]: To see all syscalls, run man syscalls

## Kernel composition
\subsection{Kernel composition}

### Essential parts of Linux kernel
* Device drivers
* Filesystem drivers
* Networking drivers
* Process management

### Device driver classes
* Character devices
* Block devices
* Network interfaces

## Device model

* **Device** is an "object". It provides some properties (platform data) and resources (IRQs, registers).
* **Driver** is a set of methods. It defines how the kernel should interact
  with the device.
* **Bus** is a common parent of *devices* and *drivers*. It implements device operations, either itself
  or by binding devices to drivers
* All devices are connected to some bus
* Bus can be physical (USB, PCI, etc.) or virtual (platform)
* Devices can be connected to more than one bus (e.g. USB controller)

## Device model

![](img/device_model.png)

## Device discovery

### Discoverable bus devices
* USB, PCI, FireWire, ...
* Created during discovery process by the bus driver

### Non-discoverable devices
* I2C, SPI, ...
* Created from device tree or during machine init

\subsection{Platform devices}
## Platform devices

* Platform device is a device that is inherently not
  discoverable - e.g. I^2C devices, SoC controllers, ...
* Platform devices are bound to platform driver by matching names
* Instantiated by code or from *Device tree*
* Should be registered early

## Case study: imx31 Lite #1

```C
// arch/arm/mach-imx/mach-mx31lite.c
static unsigned int mx31lite_pins[] = {
	/* UART1 */
	MX31_PIN_CTS1__CTS1,
	MX31_PIN_RTS1__RTS1,
	MX31_PIN_TXD1__TXD1,
	MX31_PIN_RXD1__RXD1,
	/* SPI 0 */
	MX31_PIN_CSPI1_SCLK__SCLK,
	MX31_PIN_CSPI1_MOSI__MOSI,
	MX31_PIN_CSPI1_MISO__MISO,
	MX31_PIN_CSPI1_SPI_RDY__SPI_RDY,
	MX31_PIN_CSPI1_SS0__SS0,
	MX31_PIN_CSPI1_SS1__SS1,
	...
};
```
## Case study: imx31 Lite #2

```C
/* UART */
static const struct imxuart_platform_data
	uart_pdata __initconst = {
		.flags = IMXUART_HAVE_RTSCTS,
};
/* SPI */
static const struct spi_imx_master
	spi0_pdata __initconst = {
		.chipselect	= spi0_internal_chipselect,
		.num_chipselect	= ARRAY_SIZE(spi0_internal_chipselect),
};
/* NAND */
static const struct mxc_nand_platform_data
	mx31lite_nand_board_info __initconst  = {
	.width = 1,
	.hw_ecc = 1,
};
```

## Case study: imx31 Lite #3

```C
static struct smsc911x_platform_config smsc911x_config = {
	.irq_polarity	= SMSC911X_IRQ_POLARITY_ACTIVE_LOW,
	.irq_type	= SMSC911X_IRQ_TYPE_PUSH_PULL,
	.flags		= SMSC911X_USE_16BIT,
};

static struct resource smsc911x_resources[] = {
	{
		.start		= MX31_CS4_BASE_ADDR,
		.end		= MX31_CS4_BASE_ADDR + 0x100,
		.flags		= IORESOURCE_MEM,
	}, {
		/* irq number is run-time assigned */
		.flags		= IORESOURCE_IRQ,
	},
};
```

## Case study: imx31 Lite #4

```C
static struct platform_device smsc911x_device = {
	.name		= "smsc911x",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(smsc911x_resources),
	.resource	= smsc911x_resources,
	.dev		= {
		.platform_data = &smsc911x_config,
	},
};

static struct platform_device physmap_flash_device = {
	.name   = "physmap-flash",
	.id     = 0,
	.dev    = {
		.platform_data  = &nor_flash_data,
	},
	.resource = &nor_flash_resource,
	.num_resources = 1,
};
```
## Case study: imx31 Lite #5

```C
// arch/arm/mach-imx/mach-mx31lite.c
static void __init mx31lite_init(void)
{
	imx31_soc_init();
	mxc_iomux_setup_multiple_pins(mx31lite_pins,
		ARRAY_SIZE(mx31lite_pins), "mx31lite");
	imx31_add_imx_uart0(&uart_pdata);
	imx31_add_spi_imx0(&spi0_pdata);
	/* NOR and NAND flash */
	platform_device_register(&physmap_flash_device);
	imx31_add_mxc_nand(&mx31lite_nand_board_info);
	imx31_add_spi_imx1(&spi1_pdata);

	regulator_register_fixed(0, dummy_supplies,
		ARRAY_SIZE(dummy_supplies));
}
```


## Case study: imx31 Lite #6

```C
// drivers/net/ethernet/smsc/smsc911x.c
static int smsc911x_drv_probe(struct platform_device *pdev)
{
	...
	struct smsc911x_platform_config *config = dev_get_platdata(&pdev->dev);

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "smsc911x-memory");
	if (!res)
		res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	irq = platform_get_irq(pdev, 0);
	...
```

\subsection{Device binding}
## Binding device to a driver

Binding a device to a driver is done automatically by the driver core
when:

* Driver is registered and the device already exists - `driver_attach()`
* Device is created and the driver is already registered - `device_attach()`

### Manual unbinding
* rmmod'ing the platform driver module will unbind all its devices
* Using sysfs

## Unbinding driver using sysfs

```
$ ls -l /sys/bus/usb/drivers/usb
total 0
lrwxrwxrwx 1 root root    0 Feb  2 09:23 1-1 -> 
	../../../../devices/pci0000:00/0000:00:1a.0/usb1/1-1
lrwxrwxrwx 1 root root    0 Feb  2 09:23 1-1.2 -> 
	../../../../devices/pci0000:00/0000:00:1a.0/usb1/1-1/1-1.2
lrwxrwxrwx 1 root root    0 Feb  2 09:23 1-1.6 -> 
	../../../../devices/pci0000:00/0000:00:1a.0/usb1/1-1/1-1.6
--w------- 1 root root 4096 Feb  2 09:23 bind
--w------- 1 root root 4096 Feb  2 09:23 uevent
--w------- 1 root root 4096 Feb  2 09:23 unbind

$echo -n "1-1.6" > /sys/bus/usb/drivers/usb/unbind
```

\subsection{Device files}
## Device files

* Each device may expose itself to userspace via special device files,
  stored in /dev
* Each device file is bound to driver via **major** and **minor** number
  * Major number - unique number, determines device type
  * Minor number - instances of same device between each other or sub-type
* "b" vs "c" in first column of ls -l for block devices and  
  character devices, respectively.  
  Major and minor numbers at place where file size normally appears
* For a full list of static device number allocation, see `Documentation/devices.txt`

## ls -l /dev

```sh
ls -l /dev
drwxr-xr-x  2 root  root             0 Jan 25 06:45 pts
crw-rw-rw-  1 root  root        1,   8 Jan 25 06:45 random
lrwxrwxrwx  1 root  root             4 Jan 25 06:45 rtc -> rtc0
crw-------  1 root  root      250,   0 Jan 25 06:45 rtc0
brw-rw----  1 root  disk        8,   0 Jan 25 06:45 sda
brw-rw----  1 root  disk        8,   1 Jan 25 06:50 sda1
crw-rw-rw-  1 root  tty         5,   0 Jan 28 20:34 tty
crw--w----  1 root  tty         4,   0 Jan 25 06:45 tty0
crw--w----  1 root  tty         4,   1 Jan 28 19:47 tty1
crw--w----  1 root  tty         4,  10 Jan 25 06:45 tty10

```

## devices.txt - C1

```
Memory devices
  1 = /dev/mem		Physical memory access
  2 = /dev/kmem		Kernel virtual memory access
  3 = /dev/null		Null device
  4 = /dev/port		I/O port access
  5 = /dev/zero		Null byte source
  6 = /dev/core		OBSOLETE - replaced by /proc/kcore
  7 = /dev/full		Returns ENOSPC on write
  8 = /dev/random	Nondeterministic random number gen.
  9 = /dev/urandom	Faster, less secure random number gen.
 10 = /dev/aio		Asynchronous I/O notification interface
 11 = /dev/kmsg		Writes to this come out as printk's, reads
			export the buffered printk records.
 12 = /dev/oldmem	OBSOLETE - replaced by /proc/vmcore
```

## The story of /dev

* At the beginning, device files were created statically `~ <2.4`
  * Lot of files, nodes also for devices that are not present (thousands)
  * Running out of device numbers
* Devfs `~ >2.4`
  * Dev files only for devices that are present
  * Non-standard, not persistent device names
* udev `~ >2.5`
  * userspace utility
  * dev files created upon request from kernel
  * persistent dev file naming configurable by user
* Devtmpfs `~ 2.6 - now`
  * Early tmpfs populated with device nodes
  * No userspace required to have working /dev -> Faster boot
  * udev can run on top of it as soon as userspace starts, utilizing benefits
    of both solutions

\subsection{Character devices}
## Character devices

* Most common type of device
* Device acts as a "Stream of characters"
* Examples: `serial port, I2C, SPI, /dev/random, ...`
* Must instantiate a `cdev` structure
* Must implement `file_operations`

## struct file_operations

```C
struct file_operations {
	struct module *owner;
	loff_t (*llseek) (struct file *, loff_t, int);
	ssize_t (*read) (struct file *, char __user *,...
	ssize_t (*write) (struct file *, const char __user *,...
	ssize_t (*read_iter) (struct kiocb *, struct iov_iter *);
	ssize_t (*write_iter) (struct kiocb *, struct iov_iter *);
	int (*iterate) (struct file *, struct dir_context *);
	unsigned int (*poll) (struct file *, ...
	int (*mmap) (struct file *, struct vm_area_struct *);
	int (*open) (struct inode *, struct file *);
	int (*flush) (struct file *, fl_owner_t id);
	int (*release) (struct inode *, struct file *);
	int (*fsync) (struct file *, loff_t, loff_t, ...
	int (*fasync) (int, struct file *, int);
	int (*lock) (struct file *, int, struct file_lock *);
	...
};
```

\subsection{Block devices}
## Block devices

* Device accessed by blocks
* Most often storage devices
* Examples: `mmc, hard disks...`
* Must instantiate a `gendisk` structure
* Must implement `block_device_operations`

## struct block_device_operations

```C
struct block_device_operations {
	int (*open) (struct block_device *, fmode_t);
	void (*release) (struct gendisk *, fmode_t);
	int (*rw_page)(struct block_device *, sector_t, ...
	int (*ioctl) (struct block_device *, fmode_t, ...
	int (*compat_ioctl) (struct block_device *, ...
	long (*direct_access)(struct block_device *, ...
			long);
	unsigned int (*check_events) (struct gendisk *disk,
				      unsigned int clearing);
	int (*media_changed) (struct gendisk *);
	void (*unlock_native_capacity) (struct gendisk *);
	int (*revalidate_disk) (struct gendisk *);
	int (*getgeo)(struct block_device *, struct hd_geometry *);
	void (*swap_slot_free_notify) ...
};
```

## ioctl()

* Functions defined in `file_operations` are not sufficient
  for all purposes
* ioctl() allows to pass custom data from userspace to kernel
* Used e.g. in serial driver to set line parameters (baud rate, etc.)
* Command (`cmd` parameter) should be unique, vid.  
Documentation/ioctl-number.txt

Kernel:

```C
int (*ioctl) (struct inode *inode, struct file *fl,
	unsigned int cmd, unsigned long data);
```

Userspace:

```C
int ioctl(int fd, int cmd, ...);
```

## Procfs, sysfs

### procfs - /proc
* pseudo FS that provides information about
  kernel processes and system information.
* Older one
* Does not have a strictly defined structure
* Allows all functions from `file_operations`

### sysfs - /sys
  * Another pseudo FS
  * Since 2.6.
  * Structured, uniform way to expose system information
  * Current way of exposing driver information and/or setting
    points
  * Restricted file operations

## procfs examples

* `/proc/modules` - list of loaded modules (lsmod)
* `/proc/uptime`  - system uptime (uptime)
* `/proc/version` - kernel version (uname)
* `/proc/cpuinfo` - CPU information
* `/proc/meminfo` - memory information
* `/proc/config.gz` - kernel .config used to build running kernel
* `/proc/<num>` - information about process with PID \<num\>

## sysfs examples

* `/sys/dev` - system devices (character/block)
* `/sys/bus` - system buses
* `/sys/class` - system device classes registered to kernel
* `/sys/module` - system modules (also builtin)
* `/sys/firmware` - system firmware objects
* **sysctl** - utility to manipulate `/sys` files

\subsection{KBuild}
## Kernel build system - Kbuild

Four essential building blocks:

1. `config symbols` - for conditional build of the code or to decide `y/m/n`
2. `Kconfig files` - define meta information of `config symbols` and available
   options, used by UI tools (menuconfig, xconfig, gconfig)
3. `.config file` - a database of selected `config symbols`
4. `makefiles` - a common GNU makefiles defining the build process itself

## Example - linux-xlnx/net/Kconfig

```
menu "Character devices"

source "drivers/tty/Kconfig"

config DEVMEM
	bool "/dev/mem virtual device support"
	default y
	help
	  Say Y here if you want to support the /dev/mem device.
	  The /dev/mem device is used to access areas of physical
	  memory.
	  When in doubt, say "Y".

config DEVKMEM
	bool "/dev/kmem virtual device support"
	default y
	help
	  Say Y here if you want to support the /dev/kmem device. The
	  /dev/kmem device is rarely used, but can be used for certain
	  kind of kernel debugging operations.
	  When in doubt, say "N".
```

## Makefile - linux-xlnx/net/Makefile

```makefile
#
# Makefile for the kernel character device drivers.
#
obj-y				         += mem.o random.o
obj-$(CONFIG_TTY_PRINTK)	 += ttyprintk.o
obj-y				         += misc.o
obj-$(CONFIG_ATARI_DSP56K)	 += dsp56k.o
obj-$(CONFIG_VIRTIO_CONSOLE) += virtio_console.o
obj-$(CONFIG_RAW_DRIVER)	 += raw.o
obj-$(CONFIG_SGI_SNSC)		 += snsc.o snsc_event.o
obj-$(CONFIG_MSPEC)		     += mspec.o
obj-$(CONFIG_MMTIMER)		 += mmtimer.o
obj-$(CONFIG_UV_MMTIMER)	 += uv_mmtimer.o
obj-$(CONFIG_IBM_BSR)		 += bsr.o
```

## Menuconfig

![Menuconfig - Device drivers/Character devices](img/menuconfig_module.png)

## xconfig

![bootlin.com](img/xconfig_module.png)

\subsection{Building the kernel}
## Building the kernel

* One needs to select the target architecture
  via env. variable \alert{ARCH}. It will be used for configuration
  and build.
* When left unset, kernel will be built for host's architecture
* To see all possible architectures, check `boot/arch` folder
* When cross-compiling, set the \alert{CROSS\_COMPILE} variable
  with the prefix of the toolchain
* It is advised to export these variables to the shell. Otherwise you would
  need to set them for each command separately
* When compiling for host machine, leave both variables unset

```
export ARCH=arm
export CROSS_COMPILE=arm-none-eabihf-
```

## Building the kernel - select configuration

### Use pre-defined configuration for your setup
* Default configs for most scenarios
* Check `arch/<YourArch>/configs` for configurations
  available for your architecture
* Load selected configuration
  ```sh
  make xilinx_zynq_defconfig
  ```

### Create custom configuration
* Use one of available tools
  ```sh
  make menuconfig
  # or
  make xconfig
  # or
  make gconfig
  ```

## Building the kernel - build

* Use the `-j<corenum>` parameter to run on multiple threads
  to speedup the process
* Run the kernel build
  ```sh
  make -j4
  ```
* The build will produce:
  * `vmlinux` - the uncompressed ELF kernel image
  * `arch/<YourArch>/boot/?Image` - the compressed kernel image that can boot (bzImage, zImage)
  * `arch/<YourArch>/boot/dts/*.dtb` - the device tree binaries

## Building the kernel - installation

* run `make install` to install the kernel to host system
  * usually not used in embedded development
  * requires root privileges
  * copies the kernel image, used configuration and System.map
* run `make modules_install` to install built modules to `/lib/modules`
  * set `INSTALL_MOD_PATH` variable to specify path where modules will be copied

## Building modules out-of-tree

* You can build kernel modules at any location
* The only requirement is the presence of kernel headers
* Desktop OS Distributions often provide a package
  ```sh
  apt install linux-headers-$(uname -r)
  ```
* Minimal makefile to compile kernel module `first_module.c`:
  ```makefile
  obj-m += first_module.o
  KERNEL_TREE=/lib/modules/$(shell uname -r)/build

  all:
    make -C $(KERNEL_TREE) M=$(PWD) modules
  ```

\subsection{Loading modules}
## Manual loading and unloading of a module

* **insmod** - Loads a module from a file. Does not resolve dependencies  
  `insmod first_module.ko [args]`

* **modprobe** - Loads a module from `/lib/modules/<x.y.z-arch>`. Loads dependencies first.  
  `modprobe spi-gpio`

* **rmmod** or **modprobe -r** - Unloads a module from kernel.  
  `rmmod first_module`

## Module parameters

`linux/moduleparam.h`

* Module can take arbitrary number of named parameters
* Each parameter has assigned permissions
  * S_IRUSR, S_IRUGO, ...
* Many supported types: `byte, short, ushort, int, uint, long, ulong,
  charp, bool, invbool`
* Use `modinfo` utility to show available module parameters
* Check `/usr/modules/<module>/parameters` to see all parameters,
  and get/set their value

```C
static int loops = 0;
module_param(loops, int, S_IRUGO);
MODULE_PARM_DESC(loops, "A number of loops");

static char *text = NULL;
module_param(text, charp, S_IRUGO);
MODULE_PARM_DESC(text, "This is a char pointer (string)");
```

## Automatic loading of a module

* When a device is inserted - *udev*
  * Kernel sends *uevent* to udev, udev runs appropriate actions (modprobe, create /dev node)
  * "udev rules" in /etc/udev/rules.d (allow access to device to normal users)
  * `udevadm monitor`
* During boot by specifying in `/etc/modprobe.conf`


## How to find out that there is a missing driver

* No special mechanism
* Check `dmesg`
* If applicable, check `lsusb` or `lspci`
* Find out whether required module is loaded  
  `lsmod`
* Check if the module is built into kernel  
  `cat /lib/modules/$(uname -r)/modules.builtin`

\subsection{Library}
## Library {.allowframebreaks}

[Linux source - Elixir](https://elixir.bootlin.com/linux/v5.5.1/source)  
[Kernel device drivers](http://2010.rmll.info/IMG/pdf/kernel-device-drivers-rmll2010.pdf)  
[Linux build system](https://www.linuxjournal.com/content/kbuild-linux-kernel-build-system)  
[Platform device API](https://lwn.net/Articles/448499/)  
[A fresh look at the kernel's device model](https://lwn.net/Articles/645810/)
