# Accessing physical memory

## Physical vs virtual memory

* **Physical memory** - defined by hardware, can be different for each device on the memory bus
* **Virtual memory** - as seen from software / behind MMU
* Translation (mapping) done by memory management unit (MMU)
* Smallest unit - **Page**, usually 4 kB
* Single **page frame** (physical) may be mapped multiple times (to multiple virtual pages)
* The relation between physical and virtual addresses is stored in a hierarchy of **Page tables**

## Memory mapping

![](img/virtual_memory.png)

## Motivation

* Each user process resides in its own address space
  * One process can't corrupt kernel memory
  * One process can't corrupt another process's memory
  * Each process has different virtual - physical mapping
  * More processes can map same chunk of RAM (e.g. for RPC)
* User process' memory is assigned by kernel and controlled by MMU
* Kernel memory is permanently mapped (at PAGE_OFFSET)
  * Performance
  * Handling interrupts, exceptions, syscalls, ...

\subsection{Kernel virtual memory}
## Kernel virtual memory

* `void *kmalloc(size_t size, gfp_t flags);`  
  allocate normal kernel contiguous memory
* `void *vmalloc(unsigned long size);`  
  allocate non-contiguous memory (in separate addres space).
  Usually for large allocations. Allocate entire pages.
* `void __iomem *ioremap(resource_size_t res_cookie, size_t size);`  
  Map device memory to kernel
* `void *kmap(struct page *page);`  
  Permanently map arbitrary physical page to kernel. Use `kunmap()`
  as soon as not required.

## ioremap() caching on ARM

```C
// arch/arm/include/asm/io.h
/*
 * Function		Memory type	Cacheability	Cache hint
 * ioremap()          Device		n/a		n/a
 * ioremap_nocache()  Device		n/a		n/a
 * ioremap_cache()    Normal		Writeback	Read allocate
 * ioremap_wc()       Normal		Non-cacheable	n/a
 * ioremap_wt()       Normal		Non-cacheable	n/a
 *
 * All device mappings have the following properties:
 * - no access speculation
 * - no repetition (eg, on return from an exception)
 * - number, order and size of accesses are maintained
 * - unaligned accesses are "unpredictable"
 * - writes may be delayed before they hit the endpoint device
 */
```

## /dev/mem

* Driver in `drivers/char/mem.c`
* Character device for interfacing userspace applications with physical memory
* Available operations:
  * read() - read from physical memory
  * write() - write to physical memory
  * mmap() - map physical range to userspace
* Some limitations may apply (CONFIG_STRICT_DEVMEM)

## Memory-mapped IO

```C
// drivers/gpio/gpio-zynq.c
/* Fetch the memory resource from the device */
res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
/* Map registers to kernel */
gpio->base_addr = devm_ioremap_resource(&pdev->dev, res);
if (IS_ERR(gpio->base_addr))
  return PTR_ERR(gpio->base_addr);
// ...
/* set the GPIO pin as output */
reg = readl_relaxed(gpio->base_addr + ZYNQ_GPIO_DIRM_OFFSET(bank_num));
reg |= BIT(bank_pin_num);
writel_relaxed(reg, gpio->base_addr + ZYNQ_GPIO_DIRM_OFFSET(bank_num));
```

## Memory caching on ARM

![](img/caching.png)

\subsection{Process virtual memory}
## Process virtual memory

* Show memory mapping of process with PID `<pid>`  
  `/proc/<pid>/maps`

## Library

[Memory management in Linux](https://elinux.org/images/b/b0/Introduction_to_Memory_Management_in_Linux.pdf)  
[kvmalloc()](https://lwn.net/Articles/711653/)  
[KAISER: hiding the kernel from user space](https://lwn.net/Articles/738975/)  

## Quote

> The memory management on the PowerPC can be used to frighten small children.
>
> -- <cite>Linus Torvalds</cite>

