# Inter-Process-Communication (IPC) In Linux

## IPC - Overview
* Allows to create more complex systems
* Multiple processes handling a portion of the system communicating with each other
* List of available mechanisms: **Signals**, Anonymous Pipes, **Named Pipes or FIFOs**, SysV Message Queues, POSIX Message Queues, SysV Shared memory, POSIX Shared memory, SysV semaphores, POSIX semaphores, FUTEX locks, **File-backed and anonymous shared memory using mmap**, UNIX Domain Sockets, Netlink Sockets, **Network Sockets**, Inotify mechanisms, FUSE subsystem, D-Bus subsystem, micro bus (OpenWRT)
* Availability depends on actual distribution
* Link: [Linux IPC Mechanisms](http://www.chandrashekar.info/articles/linux-system-programming/introduction-to-linux-ipc-mechanims.html)

## IPC - Signals
\subsection{Signals}
* One way asynchronous notifications
* Sent by kernel or a process to another process
* Typically alerts on an event - CTRL+C pressing, Stack fault, User defined signal...
* Signals can be: raised, caught, acted upon, ignored
* Handles signals cause execution of a signal handler function
* Acts like an interrupt - once execution ends, main context continues
* Link: [Linux Process and Signals](https://www.bogotobogo.com/Linux/linux_process_and_signals.php)

## IPC - Signals : Example

```C
#include <stdio.h>
#include <unistd.h>
#include <signal.h>

void my_signal_interrupt(int sig)
{
  printf("I got signal %d\n", sig);
  (void) signal(SIGINT, SIG_DFL);
}

int main()
{
  (void) signal(SIGINT,my_signal_interrupt);

  while(1) {
      printf("Waiting for interruption...\n");
      sleep(1);
  }
}
```


## IPC - Named Pipes
\subsection{Named Pipes}
* Acts as a FIFO
* Created by ```mkfifo test```, where 'test' is the name of the pipe
* One process can write to the file
* Other process opens the file and reads data until EOF is reached

## IPC - Named Pipes: Writer
```C
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
void main(void)
{
    int f;

    printf("Wait until a process opens FIFO for reading...\n");
    f = open("test", O_WRONLY); //open FIFO called "test"
    printf("Write the message...\n");
    write(f,"hello",5); //write 5 bytes to FIFO
    close(f);
    printf("Message Delivered!\n");
    return;
}
```

## IPC - Named Pipes: Reader
```C
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
void main(void)
{
    char buffer[32]; //some buffer
    int count;
    int f;

    printf("Wait for a process to open FIFO for writing...\n");
    f = open("test", O_RDONLY); //open FIFO called "test"
    printf("Read the message...\n");
    count = read(f, buffer, 32); //read up to 32 bytes
    close(f);
    buffer[count] = '\0';
    printf("Received: '%s'\n", buffer);
    return;
}
```

## IPC - Shared Memory
\subsection{Shared Memory}
* Allows for multiple writers and multiple readers
* Explained on a practical example - logger process
* Link: [Interprocess Communication Using Posix Shared Memory In Linux](https://www.softprayog.in/programming/interprocess-communication-using-posix-shared-memory-in-linux)
* Link: [Memory Mapped Files and Shared Memory](https://github.com/angrave/SystemProgramming/wiki/File-System%2C-Part-6%3A-Memory-mapped-files-and-Shared-memory)

## IPC - Shared Memory: Principle
* Multiple processes map the same physical memory to their respective shared memories
* Care must be taken to respect caching
* Care must be taken to provide semaphores/mutexes to avoid race conditions
* A fast replacement for read and write operations
* After the mapping, data are sent to the other process by directly writing a variable

## IPC - Shared Memory: Important C Functions
* Using `shm_open()` function a shared memory object is made available
* Using `shm_unlink()`, the object can be "closed"
* Using `ftruncate()`, the initial size of the shared memory is set
* Using `mmap()`, the shared memory is mapped to the process virtual memory
* Using `munmap()`, the mapping is removed

## IPC - Shared Memory: Logger
![](img/logger.png)

## IPC - Network Sockets
\subsection{Network Sockets}
* Using TCP or UDP communication on 127.0.0.1 / localhost
* Very common - easily portable to networked scenario
* Slower than UNIX Domain Sockets, but cross platform (Windows)
* Not file-backed - data written to a socket is lost on reboot (unlike named pipes)

