/*
 * GPIO defaults driver
 *
 * Copyright (C) 2020 Acrios systems s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/err.h>
#include <linux/gpio.h>
#include <linux/gpio/consumer.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/property.h>
#include <linux/slab.h>

#define GD_PIN_MAX 100

struct gpio_defaults_data {
	u8 gpio_cnt;
	struct gpio_desc *gpio[GD_PIN_MAX];
};

static int gpio_defaults_probe(struct platform_device *pdev)
{
	struct gpio_defaults_data *data;
	struct device *dev;
	struct fwnode_handle *child;
	struct gpio_desc *gpio;
	const char *state;

	dev = &pdev->dev;

	data = devm_kzalloc(dev, sizeof(struct gpio_defaults_data),
		GFP_KERNEL);

	/* Iterate over all child nodes */
	device_for_each_child_node(dev, child) {
		// struct device_node *np = to_of_node(child);

		gpio = devm_get_gpiod_from_child(dev, NULL, child);
		if (IS_ERR(gpio)) {
			fwnode_handle_put(child);
			return ERR_CAST(gpio);
		}

		if (!fwnode_property_read_string(child, "default-state",
						 &state)) {
			if (!strcmp(state, "on")) {
				gpiod_direction_output(gpio, 1);
			} else if (!strcmp(state, "off")) {
				gpiod_direction_output(gpio, 0);
			}
		}
		data->gpio[data->gpio_cnt++] = gpio;
	}

	dev_info(dev, "Set default state of %d pins\n", data->gpio_cnt);

	platform_set_drvdata(pdev, data);

	return 0;
}

static void gpio_defaults_shutdown(struct platform_device *pdev)
{
	struct gpio_defaults_data *data = platform_get_drvdata(pdev);
	int i;

	dev_info(&pdev->dev, "Reset default state of %d pins\n", data->gpio_cnt);

	for (i = 0; i < data->gpio_cnt; i++) {
		gpiod_direction_input(data->gpio[i]);
	}
}

static const struct of_device_id of_gpio_defaults_match[] = {
	{ .compatible = "gpio-defaults", },
	{},
};

MODULE_DEVICE_TABLE(of, of_gpio_defaults_match);

static struct platform_driver gpio_defaults_driver = {
	.probe		= gpio_defaults_probe,
	.shutdown	= gpio_defaults_shutdown,
	.driver		= {
		.name	= "defaults-gpio",
		.of_match_table = of_gpio_defaults_match,
	},
};

module_platform_driver(gpio_defaults_driver);

MODULE_AUTHOR("Lukas Janik <janik@acrios.com>");
MODULE_DESCRIPTION("GPIO Defaults driver");
MODULE_LICENSE("GPL");
